


from fqllang.fql_stage.analizers.create_form import CreateFormAnalizer
from fqllang.fql_stage.utils import getClassName
from fqllang.sql_stage.code_transpilers.transpiler import FqlAstSqlTranspiler
from fqllang.sql_stage.code_generators.create import CreatePostgreSql
from fqllang.sql_stage.code_generators.insert import \
    PreparedInsertPostgreSql
from fqllang.sql_stage.code_generators.select import (
    SelectPostgreSql, SelectWithCriteriaPostgreSql)
from fqllang.sql_stage.code_generators.basic_elements import (
    ColumnList, Condition, Criteria, DataConstraint, DataType, Field,
    FieldList, Reference, ValueList, convertFqlDataTypeToPostgreSqlDataType)


class FqlAstPostgreSqlTranspiler(FqlAstSqlTranspiler):
    def __init__(self, fqlAstModel) -> None:
        super().__init__(fqlAstModel)

    def generate(self):
        return self.fqlStatementTranspiler(self.model)

    def __str__(self) -> str:
        return self.generate()

    def fqlStatementTranspiler(self, model):
        className = getClassName(model)
        if className == "CreateForm":
            form = CreateFormAnalizer(model).make_copy()
            return form.sql_code()
            # return self.createFormTranspiler(model).generate()
        elif className == "CreateNew":
            return self.createNewTranspiler(model).generate()
        elif className == "GetCase":
            return self.getCaseTranspiler(model).generate()
        raise Exception("Modelo incorrecto")

    def createFormTranspiler(self, createForm):
        formName = createForm.formName

        dataSpecList = self.dataSpecListTranspiler(createForm.dataSpecList)

        keyField = Field('FQL_ID', DataType.serial, DataConstraint.primaryKey)

        fieldList = FieldList(keyField, *dataSpecList)

        return CreatePostgreSql(formName, ColumnList.emptyColumnList(), fieldList)

    def dataSpecListTranspiler(self, dataSpecList):
        dataSpecListTranspiled = map(lambda dataSpec: self.dataSpecTranspiler(
            dataSpec), dataSpecList)
        return list(dataSpecListTranspiled)

    def dataSpecTranspiler(self, dataSpec):
        className = getClassName(dataSpec)
        if className == "DataDefinition":
            return self.dataDefinitionTranspiler(dataSpec)
        elif className == "DataReference":
            return self.dataReferenceTranspiler(dataSpec)

    def dataDefinitionTranspiler(self, dataDefinition):
        dataName = dataDefinition.dataName
        dataType = convertFqlDataTypeToPostgreSqlDataType(
            dataDefinition.dataType)
        notNull = dataDefinition.notNull
        unique = dataDefinition.unique
        field = Field(dataName, dataType)
        if notNull:
            field.addDataConstraint(DataType.notNull)
        if unique:
            field.addDataConstraint(DataType.unique)
        return field

    def dataReferenceTranspiler(self, dataReference):
        dataName = dataReference.dataName
        refCardinality = dataReference.refCardinality
        refPath = dataReference.refPath
        basicField = Field(dataName, DataType.integer)
        reference = Reference(basicField, refPath)
        return reference

    def createNewTranspiler(self, createNew):
        formName = createNew.formName
        columns, values = self.createNewValuesTranspiler(
            createNew.createNewValues)
        return PreparedInsertPostgreSql(formName, columns, values)

    def createNewValuesTranspiler(self, createNewValues):
        keyValueEntryList = createNewValues.keyValueEntryList
        keyValueList = [self.keyValueEntryTranspiler(
            keyValueEntry) for keyValueEntry in keyValueEntryList]
        columns = []
        values = []
        for key, value in keyValueList:
            columns.append(key)
            values.append(value)
        return ColumnList.createColumnListByElements(*columns),\
                ValueList.createColumnListByElements(*values)

    def keyValueEntryTranspiler(self, keyValueEntry):
        className = getClassName(keyValueEntry)
        if className == "KeyValueSimple":
            return self.keyValueSimpleTranspiler(keyValueEntry)
        elif className == "KeysSeparatesWithDots":
            pass

    def keyValueSimpleTranspiler(self, keyValueSimple):
        key = keyValueSimple.key
        valueTerminal = keyValueSimple.valueTerminal
        return key, valueTerminal

    def getCaseTranspiler(self, getCase):
        formName = getCase.formName
        conditions = self.extractConditions(getCase)
        if conditions:
            criteria = self.getCaseConditionsTranspiler(conditions)
            return SelectWithCriteriaPostgreSql(formName, ColumnList.emptyColumnList(), criteria)
        else:
            return SelectPostgreSql(formName)

    def getCaseConditionsTranspiler(self, conditions):
        conditionList = [self.getCondition(condition)
                         for condition in conditions]
        return Criteria(*conditionList)

    def getCondition(self, condition):
        className = getClassName(condition)
        if className == "ConditionTerminal":
            attribute = condition.attribute
            valueTerminal = condition.valueTerminal
        return Condition(attribute, valueTerminal)

    def extractConditions(self, getCase):
        return getCase.conditions




from fqllang.sql_stage.code_generators.create import CreatePostgreSql
from fqllang.sql_stage.code_generators.basic_elements import ColumnList, DataConstraint, DataType, Field, FieldList


class TestCreatePostgreSql:

    def test_generate(self):
        tableName = "Person"
        columnList = ColumnList.createColumnListByElements('name', 'age')
        fieldList = FieldList(Field(
            "name", DataType.varchar, DataConstraint.notNull), Field("age", DataType.integer))
        create = CreatePostgreSql(tableName, columnList, fieldList)
        assert create.generate() == "CREATE TABLE Person(name VARCHAR(511) NOT NULL,age INTEGER);"



from fqllang.sql_stage.code_generators.basic_elements import Column, ColumnList, Condition, Criteria, DataConstraint, DataType, Field, FieldList, Reference, ValueList, Where


class TestColumnList:
    def test_len(self):
        name = Column('name')
        age = Column('age')
        columnList = ColumnList(name, age)
        assert len(columnList) == 2

    def test_isEmpty(self):
        columnList = ColumnList()
        assert columnList.isEmpty

    def test_transformColumnsInStrList(self):
        name = Column('name')
        age = Column('age')
        columnList = ColumnList(name, age)
        assert columnList.columnsInStrList() == ['name', 'age']

    def test_iter(self):
        name = Column('name')
        age = Column('age')
        columnList = ColumnList(name, age)
        for column, element in zip(columnList, [name, age]):
            assert column == element

    def test_generate(self):
        name = Column('name')
        age = Column('age')
        columnList = ColumnList(name, age)
        assert columnList.generate() == "(name,age)"

    def test_generateEmpty(self):
        columnList = ColumnList()
        assert columnList.generate() == "*"

    def test_createColumnListByElements(self):
        strElements = ["name", "age", "address"]
        columnList = ColumnList.createColumnListByElements(*strElements)
        assert columnList.generate() == "(name,age,address)"
        
class TestValueList:
    def test_generate(self):
        valueList = ValueList.createColumnListByElements(4,"Leandro",True)
        assert valueList.generate() == "(4,'Leandro',True)"


class TestField:
    def test_generate(self):
        name = "name"
        dataType = DataType.varchar
        dataConstraint1 = DataConstraint.notNull
        dataConstraint2 = DataConstraint.unique
        field = Field(name, dataType, dataConstraint1, dataConstraint2)
        assert field.generate() == "name VARCHAR(511) NOT NULL UNIQUE"

    def test_generateWithoutDataConstraints(self):
        name = "name"
        dataType = DataType.varchar
        field = Field(name, dataType)
        assert field.generate() == "name VARCHAR(511)"

    def test_addDataConstraint(self):
        name = "name"
        dataType = DataType.varchar
        field = Field(name, dataType)
        field.addDataConstraint(DataConstraint.primaryKey)
        field.addDataConstraint(DataConstraint.primaryKey)
        assert field.generate() == "name VARCHAR(511) PRIMARY KEY"


class TestCondition:
    def test_generate(self):
        condition1 = Condition('name', 'Leandro')
        condition2 = Condition('age', 20)
        assert condition1.generate() == "name='Leandro'"
        assert condition2.generate() == "age=20"


class TestCriteria:
    def test_generate(self):
        condition1 = Condition('name', 'Leandro')
        condition2 = Condition('age', 20)
        criteria = Criteria(condition1, condition2)
        assert criteria.generate() == "name='Leandro' and age=20"


class TestWhere:
    def test_generate(self):
        condition1 = Condition('name', 'Leandro')
        condition2 = Condition('age', 20)
        criteria = Criteria(condition1, condition2)
        where = Where(criteria)
        assert where.generate() == "WHERE name='Leandro' and age=20"
        

class TestReference:
    def test_generate(self):
        field = Field("Project", DataType.integer)
        refPath = "Projects"
        reference = Reference(field, refPath)
        assert reference.generate() == "Project INTEGER,FOREIGN KEY(Project) REFERENCES Projects(FQL_ID)"


class TestFieldList:
    def test_generate(self):
        field1 = Field("name", DataType.varchar, DataConstraint.notNull)
        field2 = Field("age", DataType.integer)
        field3 = Field("isCold", DataType.boolean)
        fieldList = FieldList(field1, field2, field3)
        assert fieldList.generate() == "name VARCHAR(511) NOT NULL,age INTEGER,isCold BOOLEAN"
        
    def test_generateWithReferences(self):
        field1 = Field("name", DataType.varchar, DataConstraint.notNull)
        
        fieldReference = Field("Project", DataType.integer)
        refPath = "Projects"
        reference = Reference(fieldReference, refPath)
        
        fieldList = FieldList(field1, reference)
        result = "name VARCHAR(511) NOT NULL,Project INTEGER,FOREIGN KEY(Project) REFERENCES Projects(FQL_ID)"
        assert fieldList.generate() == result
        

    def test_fieldsInStrList(self):
        field1 = Field("name", DataType.varchar, DataConstraint.notNull)
        field2 = Field("age", DataType.integer)
        field3 = Field("isCold", DataType.boolean)
        fieldList = FieldList(field1, field2, field3)
        assert fieldList.fieldsInStrList(
        ) == ["name VARCHAR(511) NOT NULL", "age INTEGER", "isCold BOOLEAN"]

    def test_isEmpty(self):
        fieldList = FieldList()
        assert fieldList.isEmpty

    def test_len(self):
        field1 = Field("name", DataType.varchar, DataConstraint.notNull)
        field2 = Field("age", DataType.integer)
        field3 = Field("isCold", DataType.boolean)
        fieldList = FieldList(field1, field2, field3)
        assert len(fieldList) == 3




import fqllang.fql_stage.components.create_form as create
from fqllang.fql_stage.components.datatypes import FqlNumberDatatype
from fqllang.fql_stage.components.refcardinalities import FqlManyRefcardinality, FqlOneRefcardinality
from fqllang.sql_stage.components.basic_components import SqlField, SqlForeignKeyField
from fqllang.sql_stage.components.datatypes import SqlIntegerDatatype


class Fql:
    pass


class FqlField:
    def __init__(self, name, datatype, *dataconstraints) -> None:
        self.name = name
        self.datatype = datatype
        self.dataconstraints = dataconstraints

    def sql_code(self):
        return self.to_sql_field().generate()
    
    def to_sql_field(self):
        dataconstraints = [
            dataconstraint.__sql__ for dataconstraint in self.dataconstraints]
        datatype = self.datatype.__sql__
        return SqlField(self.name, datatype, *dataconstraints)


class FqlReference:
    # def __init__(self, dataName, refCardinality, refPath, refConstraint) -> None:
    def __init__(self, dataName, refCardinality, refPath) -> None:
        self.dataName = dataName
        self.refCardinality = refCardinality
        self.refPath = refPath
        # self.refConstraint = refConstraint        

    def get_fields(self):
        def create_fields():
            key = FqlField(self.dataName, FqlNumberDatatype)
            foreignKey = FqlForeignKey(self.dataName, self.refPath)
            return [key, foreignKey]
        return [] if self.refCardinality.__type__=='many' else create_fields()

    def get_extracomponent(self, baseName):
        def create_form(baseName):
            name = f"{baseName}_{self.refPath}"
            basename_field = FqlField(f"fk_{baseName}", FqlNumberDatatype)
            fk_basename = FqlForeignKey(f"fk_{baseName}", baseName)
            refpath_field = FqlField(f"fk_{self.refPath}", FqlNumberDatatype)
            fk_refpath = FqlForeignKey(f"fk_{self.refPath}", self.refPath)
            return create.Form(name, basename_field, fk_basename, refpath_field, fk_refpath)
        if self.refCardinality.__type__=='many': return create_form(baseName)
    
class FqlForeignKey:
    def __init__(self, key, refpath) -> None:
        self.key = key
        self.refpath = refpath
        
    def sql_code(self):
        return self.to_sql_reference().generate()
    
    def to_sql_reference(self):
        return SqlForeignKeyField(self.key, self.refpath, "FQL_ID")


class RefCardinality:
    def __init__(self, cardinality1=None, cardinality2=None) -> None:
        self.cardinality1 = cardinality1
        self.cardinality2 = cardinality2


class RefConstraint:
    def __init__(self, constraint) -> None:
        self.constraint = constraint


class SimpleId:
    def __init__(self, id) -> None:
        self.id = id
        
    def sql_code(self):
        return str(self.id)
    
    def __str__(self) -> str:
        return self.sql_code()


class DataPath:
    def __init__(self, *dataPath) -> None:
        self.path = dataPath

    def __iter__(self):
        for data in self.path:
            yield data


class DataLabelList:
    def __init__(self, *dataLabels) -> None:
        self.dataLabels = dataLabels


class DataLabel(SimpleId):
    def __init__(self, id) -> None:
        super().__init__(id)


class All:
    def __init__(self, all) -> None:
        self.star = all



class CreateTable:
    def __init__(self, name, *fields) -> None:
        self.name = name
        self.fields = fields

    def generate(self):
        result = self.generate_header()
        result += self.generate_body()
        result += self.generate_footer()
        return result

    def generate_header(self):
        return f"create table {self.name}("

    def generate_body(self):
        return ",".join(field.generate() for field in self.fields)

    def generate_footer(self):
        return ");"
    
    def str(self):
        return self.generate()



from fqllang.fql_stage.components.basic_components import FqlField, FqlForeignKey, FqlReference
from fqllang.fql_stage.components.create_form import Form
from fqllang.fql_stage.components.dataconstraints import FqlUniqueDataConstraint
from fqllang.fql_stage.components.datatypes import FqlTextDatatype
from fqllang.fql_stage.components.refcardinalities import FqlManyRefcardinality, FqlOneRefcardinality


class TestFqlField:
    def test_sql_code(self):
        name = "name"
        datatype = FqlTextDatatype
        dataconstraint1 = FqlUniqueDataConstraint
        field = FqlField(name, datatype, dataconstraint1)
        assert field.sql_code() == "name varchar(511) unique"
        
class TestFqlForeignKey:
    def test_sql_code(self):
        foreign = FqlForeignKey("Project", "Projects")
        assert foreign.sql_code() == "foreign key(Project) references Projects(FQL_ID)"
        
class TestFqlReference:
        
    def test_get_fields_empty(self):
        name = "Project"
        refcard = FqlManyRefcardinality
        refpath = "Projects"
        reference = FqlReference(name, refcard, refpath)
        assert len(reference.get_fields())==0
        
    def test_get_fields(self):
        name = "Project"
        refcard = FqlOneRefcardinality
        refpath = "Projects"
        reference = FqlReference(name, refcard, refpath)
        assert len(reference.get_fields())==2
        
    def test_get_extracomponent(self):
        name = "Project"
        refcard = FqlManyRefcardinality
        refpath = "Projects"
        component = FqlReference(name, refcard, refpath).get_extracomponent("Employees")
        assert isinstance(component, Form)





from itertools import chain

import fqllang.fql_stage.components.basic_components as components
from fqllang.fql_stage.components.dataconstraints import FqlPrimarykeyDataConstraint
from fqllang.fql_stage.components.datatypes import FqlSerialDatatype
from fqllang.sql_stage.components.create_table import CreateTable


class Form:
    def __init__(self, name, *fields) -> None:
        self.name = name
        hidden_field = components.FqlField("FQL_ID", FqlSerialDatatype, FqlPrimarykeyDataConstraint)
        self.fields = [hidden_field, *fields]

    def sql_code(self):
        subfields, extracomponents = self.classify_fields()
        sqlsubfields = self.to_sql_subfields(subfields)
        result = CreateTable(self.name, *sqlsubfields).generate()
        result += self.generate_extracomponents(extracomponents)
        return result

    def classify_fields(self):
        subfields = filter(lambda field: isinstance(field, components.FqlField), self.fields)
        references = filter(lambda field: isinstance(field, components.FqlReference), self.fields)
        extrasubfields, extracomponents = self.classify_references(list(references))
        subfields = list(chain(subfields,extrasubfields))
        return subfields, extracomponents
    
    def classify_references(self, references):
        subfields = map(lambda reference: reference.get_fields(),references)
        components = map(lambda reference: reference.get_extracomponent(self.name), references)
        subfields = filter(lambda subfield: len(subfield)!=0,subfields)
        components = filter(lambda component: component is not None,components)
        subfields = chain(*subfields)
        return list(subfields), list(components)
    
    def to_sql_subfields(self, subfields):
        result = []
        for subfield in subfields:
            if isinstance(subfield, components.FqlField):
                result.append(subfield.to_sql_field())
            elif isinstance(subfield, components.FqlForeignKey):
                result.append(subfield.to_sql_reference())
        return result

    def generate_extracomponents(self, extracomponents):
        components = []
        for component in extracomponents:
            if isinstance(component, Form):
                subfields = self.to_sql_subfields(component.fields)
                components.append(CreateTable(component.name, *subfields))
        return "\n" + "\n".join(component.generate() for component in components)



class SqlDatatype:
    __type__="default"

class SqlIntegerDatatype(SqlDatatype):
    __type__="integer"
    
class SqlVarcharDatatype(SqlDatatype):
    __type__="varchar(511)"
    
class SqlSerialDatatype(SqlDatatype):
    __type__="serial"
    
class SqlBooleanDatatype(SqlDatatype):
    __type__="boolean"
    



from fqllang.fql_stage.ast_builders import FqlAstBuilder


class TestFqlAstBuilder:
    def test_fqlAstFromCode_createForm(self):
        code = "create form Person(name TEXT, age NUMBER)"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        assert fqlAstBuilder.model

    def test_fqlAstFromCode_createNew_withNew(self):
        code = "create new Person values (1 , 'Leandro', 'Hernandez')"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        assert fqlAstBuilder.model

    def test_fqlAstFromCode_createNew_WithoutNew(self):
        code = "create Person values (1 , 'Leandro', 'Hernandez')"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        assert fqlAstBuilder.model

    def test_fqlAstFromCode_createNew_keyValues(self):
        code = "create Person (number=1 , name='Leandro', H.G.FQL_ID=null, A.FQL_VERSION=2)"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        assert fqlAstBuilder.model

    def test_fqlAstFromCode_createNew_newkeyValues(self):
        code = "create new Person (number=1 , name='Leandro', H.G.FQL_ID=null, A.FQL_VERSION=2)"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        assert fqlAstBuilder.model

    def test_fqlAstFromCode_showForms(self):
        code = "show forms Person"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        assert fqlAstBuilder.model

    def test_fqlAstFromCode_getCase_retrieveAllFormAttributes(self):
        code = "get Person"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        assert fqlAstBuilder.model

    def test_fqlAstFromCode_getCase_withConditions(self):
        code = "get Person with Person.name=5 and age=24"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        assert fqlAstBuilder.model

    def test_fqlAstFromCode_getCase_withSomeAttributes(self):
        code = "get Person (code, firstName, age)"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        assert fqlAstBuilder.model

    def test_fqlAstFromCode_getCase_withConditionsAndSomeAttributes(self):
        code = "get Person (code, firstName, age) with Person.age=1 and firstName='Leandro'"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        assert fqlAstBuilder.model





from fqllang.fql_stage.components.basic_components import Fql


class CreateNew(Fql):
    def __init__(self, formName, *newValues) -> None:
        super().__init__()
        self.formName = formName
        self.newValues = newValues
        
class CreateNewValues(Fql):
    def __init__(self, keys, values) -> None:
        super().__init__()
        self.keys = keys
        self.values = values
    
    @staticmethod
    def createNewValuesFromValueList(*values):
        return CreateNewValues(None,*values)
        




from fqllang.sql_stage.components.basic_components import SqlField, SqlForeignKeyField
from fqllang.sql_stage.components.create_table import CreateTable
from fqllang.sql_stage.components.datatypes import SqlIntegerDatatype, SqlVarcharDatatype


class TestCreateTable:
    def test_generate(self):
        field1 = SqlField("name", SqlVarcharDatatype)
        field2 = SqlField("age", SqlIntegerDatatype)
        table = CreateTable("Project", field1, field2)
        assert table.generate() == "create table Project(name varchar(511),age integer);"
        
    def test_generate_with_foreignkey(self):
        field1 = SqlField("name", SqlVarcharDatatype)
        field2 = SqlField("Project", SqlIntegerDatatype)
        foreignkey = SqlForeignKeyField("Project", "Projects", "FQL_ID")
        table = CreateTable("Person", field1, field2, foreignkey)
        assert table.generate() == "create table Person(name varchar(511),Project integer,foreign key(Project) references Projects(FQL_ID));"



from enums import Enum

DataType = Enum("DataType", integer="INTEGER",
                varchar="VARCHAR(511)", boolean="BOOLEAN", serial="SERIAL")

PostgreSqlDataType = Enum("DataType", integer="INTEGER",
                          varchar="VARCHAR(511)", boolean="BOOLEAN", serial="SERIAL")

FqlDataType = Enum("FqlDataType", number="NUMBER",
                   text="TEXT", boolean="BOOLEAN")

DataConstraint = Enum("DataConstraint", notNull="NOT NULL",
                      unique="UNIQUE", primaryKey="PRIMARY KEY")


def convertFqlDataTypeToPostgreSqlDataType(fqlDataType):
    try:
        fql_sql_dict = {
            "NUMBER": PostgreSqlDataType.integer,
            "TEXT": PostgreSqlDataType.varchar,
            "BOOLEAN": PostgreSqlDataType.boolean
        }
        return fql_sql_dict[fqlDataType]
    except Exception as e:
        raise e


class Condition:
    def __init__(self, attribute, value) -> None:
        self.attribute = attribute
        if isinstance(value, str):
            self.value = f"'{value}'"
        else:
            self.value = value

    def generate(self):
        return f"{self.attribute}={self.value}"

    def __str__(self) -> str:
        return self.generate()


class Criteria:
    def __init__(self, *conditions) -> None:
        self.conditions = conditions

    def generate(self):
        conditionsToStrList = self.transformConditionsToStrList()
        return ' and '.join(conditionsToStrList)

    def transformConditionsToStrList(self):
        result = []
        for condition in self.conditions:
            result.append(condition.generate())
        return result

    def __str__(self) -> str:
        return self.generate()


class Where:
    def __init__(self, criteria: Criteria) -> None:
        self.criteria = criteria

    def generate(self):
        return f"WHERE {self.criteria}"

    def __str__(self) -> str:
        return self.generate()


class Field:
    def __init__(self, name: str, dataType, *dataConstraints):
        self.name = name
        self.dataType = dataType.value
        self.dataConstraints = list(map(lambda x: x.value, dataConstraints))

    def addDataConstraint(self, dataConstraint):
        if not self.existDataConstraint(dataConstraint):
            self.dataConstraints.append(dataConstraint.value)

    def existDataConstraint(self, dataConstraint):
        dataConstraintValue = dataConstraint.value
        for value in self.dataConstraints:
            if value == dataConstraintValue:
                return True
        return False

    def generate(self):
        fieldElements = [self.name, self.dataType, *self.dataConstraints]
        result = " ".join(fieldElements)
        return result

    def __str__(self):
        return self.generate()


class Reference:
    def __init__(self, field: Field, refPath: str) -> None:
        self.field = field
        self.refPath = refPath

    def generate(self):
        fieldStr = self.field.generate()
        referenceStr = self.generateReferenceStr()
        return f"{fieldStr},{referenceStr}"

    def generateReferenceStr(self):
        fieldName = self.field.name
        fqlId = "FQL_ID"
        refPath = self.refPath
        return f"FOREIGN KEY({fieldName}) REFERENCES {refPath}({fqlId})"


class FieldList:
    def __init__(self, *fields) -> None:
        self.fields = fields

    def generate(self):
        fieldElementsStrList = self.fieldsInStrList()
        result = ",".join(fieldElementsStrList)
        return result

    def fieldsInStrList(self):
        fieldElementsStrList = map(lambda x: x.generate(), self)
        return list(fieldElementsStrList)

    def addField(self, field):
        self.fields.append(field)

    def __iter__(self):
        yield from self.fields

    @property
    def isEmpty(self):
        return len(self) == 0

    def __len__(self):
        return len(self.fields)

    def __str__(self) -> str:
        return self.generate()


class Column:
    def __init__(self, value) -> None:
        self.value = value

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value


class ColumnList:
    def __init__(self, *columns) -> None:
        self.columns = columns

    @staticmethod
    def createColumnListByElements(*elements):
        columns = map(Column, elements)
        return ColumnList(*columns)

    @staticmethod
    def emptyColumnList():
        return ColumnList()

    def generate(self):
        if self.isEmpty:
            return "*"
        columnsInStr = self.columnsInStrList()
        result = "(" + ",".join(columnsInStr) + ")"
        return result

    @property
    def isEmpty(self):
        return len(self) == 0

    def __len__(self):
        return len(self.columns)

    def columnsInStrList(self):
        result = map(str, self)
        return list(result)

    def __iter__(self):
        yield from self.columns

    def __str__(self) -> str:
        return self.generate()


class ValueList(ColumnList):
    def __init__(self, *values) -> None:
        super().__init__(*values)

    @staticmethod
    def createColumnListByElements(*elements):
        def valueInStr(value):
            if isinstance(value, str):
                return f"'{value}'"
            return str(value)
        columns = map(lambda element: Column(valueInStr(element)), elements)
        return ColumnList(*columns)



from fqllang.sql_stage.components.datatypes import SqlBooleanDatatype, SqlDatatype, SqlIntegerDatatype, SqlSerialDatatype, SqlVarcharDatatype


class FqlDatatype:
    __type__ = "default"
    __sql__ = SqlDatatype


class FqlNumberDatatype(FqlDatatype):
    __type__ = "number"
    __sql__ = SqlIntegerDatatype


class FqlTextDatatype(FqlDatatype):
    __type__ = "text"
    __sql__ = SqlVarcharDatatype


class FqlSerialDatatype(FqlDatatype):
    __type__ = "serial"
    __sql__ = SqlSerialDatatype


class FqlBooleanDatatype(FqlDatatype):
    __type__ = "boolean"
    __sql__ = SqlBooleanDatatype





from fqllang.fql_stage.components.basic_components import FqlField, FqlReference
from fqllang.fql_stage.components.create_form import Form
from fqllang.fql_stage.components.datatypes import FqlTextDatatype
from fqllang.fql_stage.components.refcardinalities import FqlManyRefcardinality, FqlOneRefcardinality


class TestForm:
    def test_classify_references(self):
        name = "Projects"
        field1 = FqlField("name", FqlTextDatatype)
        reference1 = FqlReference(
            "Employee", FqlManyRefcardinality, "Employees")
        fields, components = Form(
            name, field1, reference1).classify_references([reference1])
        assert len(list(components)) == 1 and len(list(fields)) == 0

    def test_classify_references2(self):
        name = "Projects"
        field1 = FqlField("name", FqlTextDatatype)
        reference1 = FqlReference(
            "Employee", FqlOneRefcardinality, "Employees")
        fields, components = Form(
            name, field1, reference1).classify_references([reference1])
        assert len(list(components)) == 0 and len(list(fields)) == 2

    def test_classify_fields(self):
        name = "Projects"
        field1 = FqlField("name", FqlTextDatatype)
        reference1 = FqlReference(
            "Employee", FqlManyRefcardinality, "Employees")
        fields, components = Form(name, field1, reference1).classify_fields()
        assert len(fields) == 2 and len(components) == 1

    def test_classify_fields2(self):
        name = "Projects"
        field1 = FqlField("name", FqlTextDatatype)
        reference1 = FqlReference(
            "Employee", FqlOneRefcardinality, "Employees")
        fields, components = Form(name, field1, reference1).classify_fields()
        assert len(fields) == 4 and len(components) == 0

    def test_sql_code(self):
        name = "Projects"
        field1 = FqlField("name", FqlTextDatatype)
        reference1 = FqlReference(
            "Employee", FqlOneRefcardinality, "Employees")
        form = Form(name, field1, reference1)
        assert form.sql_code() == "create table Projects(FQL_ID serial primary key,name varchar(511),Employee integer,foreign key(Employee) references Employees(FQL_ID));\n"

    def test_sql_code2(self):
        name = "Projects"
        field1 = FqlField("name", FqlTextDatatype)
        reference1 = FqlReference(
            "Employee", FqlManyRefcardinality, "Employees")
        form = Form(name, field1, reference1)
        code1 = "create table Projects(FQL_ID serial primary key,name varchar(511));"
        code2 = "create table Projects_Employees(FQL_ID serial primary key,fk_Projects integer,foreign key(fk_Projects) references Projects(FQL_ID),fk_Employees integer,foreign key(fk_Employees) references Employees(FQL_ID));"
        assert form.sql_code() == code1 + "\n" + code2





from fqllang.fql_stage.analizers.base import Analizer
from fqllang.fql_stage.components.create_form import Form
from fqllang.fql_stage.components.dataconstraints import FqlNotnullDataConstraint, FqlUniqueDataConstraint
from fqllang.fql_stage.components.datatypes import FqlBooleanDatatype, FqlNumberDatatype, FqlTextDatatype
from fqllang.fql_stage.components.refcardinalities import FqlManyRefcardinality, FqlOneRefcardinality
from fqllang.fql_stage.utils import getClassName
from fqllang.fql_stage.components.basic_components import FqlField, FqlReference, RefCardinality, RefConstraint, SimpleId, DataPath, All, DataLabelList, DataLabel



class CreateFormAnalizer(Analizer):
    def make_copy(self):
        return Form(self.form_name, *self.data_specs)

    @property
    def data_specs(self):
        return DataSpecListAnalizer(self.model.dataSpecList).make_copy()

    @property
    def form_name(self):
        return self.model.formName


class DataSpecListAnalizer(Analizer):
    def make_copy(self):
        return self.data_specs

    @property
    def data_specs(self):
        return [self._analize_dataSpec(dataSpec) for dataSpec in self]

    def _analize_dataSpec(self, dataSpec):
        return DataSpecAnalizer(dataSpec).make_copy()

    def __iter__(self):
        yield from self.model


class DataSpecAnalizer(Analizer):
    def make_copy(self):
        className = getClassName(self.model)
        analizer = self.dataSpecDict
        return analizer[className](self.model).make_copy()

    @property
    def dataSpecDict(self):
        return {
            "DataDefinition": DataDefinitionAnalizer,
            "DataReference": DataReferenceAnalizer
        }


class DataDefinitionAnalizer(Analizer):
    def make_copy(self):
        return FqlField(self.data_name, self.data_type, *self.dataconstraints)

    @property
    def data_name(self):
        return self.model.dataName

    @property
    def data_type(self):
        def avaliable_datatypes():
            return {
                "NUMBER": FqlNumberDatatype,
                "TEXT": FqlTextDatatype,
                "BOOLEAN":FqlBooleanDatatype
            }
        datatype = self.model.dataType
        return avaliable_datatypes()[datatype]
    
    @property
    def dataconstraints(self):
        dataconstraints = []
        if self.not_null: dataconstraints.append(FqlNotnullDataConstraint)
        if self.unique: dataconstraints.append(FqlUniqueDataConstraint)
        return dataconstraints

    @property
    def not_null(self):
        return self.model.notNull

    @property
    def unique(self):
        return self.model.unique


class DataReferenceAnalizer(Analizer):
    def make_copy(self):
        # return FqlReference(self.data_name, self.ref_cardinality, self.ref_path, self.ref_constraint)
        return FqlReference(self.data_name, self.ref_cardinality, self.ref_path)

    @property
    def data_name(self):
        return self.model.dataName

    @property
    def ref_cardinality(self):
        return RefCardinalityAnalizer(self.model.refCardinality).make_copy()

    @property
    def ref_path(self):
        return RefPathAnalizer(self.model.refPath).make_copy()

    # @property
    # def ref_constraint(self):
    #     return RefConstraintAnalizer(self.model.refConstraint).make_copy()


class RefCardinalityAnalizer(Analizer):

    def make_copy(self):
        if self.model is None: return FqlOneRefcardinality
        if self.card2 is None: return FqlOneRefcardinality
        if self.card2 == "many": return FqlManyRefcardinality
        return FqlOneRefcardinality

    @property
    def card1(self):
        return self.model.cardinality1

    @property
    def card2(self):
        return self.model.cardinality2


class RefPathAnalizer(Analizer):
    def make_copy(self):
        className = getClassName(self.model)
        analizer = self.refPathDict
        return analizer[className](self.model).make_copy()

    @property
    def refPathDict(self):
        return {
            # "IdDotDataPath": IdDotDataPathAnalizer,
            "str": SimpleIdAnalizer
        }


class SimpleIdAnalizer(Analizer):
    def make_copy(self):
        return SimpleId(self.model)


# class IdDotDataPathAnalizer(Analizer):
#     def make_copy(self):
#         return DataPath(*self.full_datapath)

#     @property
#     def full_datapath(self):
#         return [*self]

#     @property
#     def id(self):
#         return self.model.id

#     @property
#     def data_path(self):
#         return DataPathAnalizer(self.model.dataPath).make_copy()

#     def __iter__(self):
#         yield self.id
#         data_path = self.data_path
#         if isinstance(data_path, list):
#             yield from data_path
#         else:
#             yield data_path


# class DataPathAnalizer(Analizer):
#     def make_copy(self):
#         className = getClassName(self.model)
#         analizer = self.dataPathDict
#         return analizer[className](self.model).make_copy()

#     @property
#     def dataPathDict(self):
#         return {
#             "IdDotDataPath": IdDotDataPathAnalizer,
#             "DataLabelList": DataLabelListAnalizer,
#             "All": StarAnalizer,
#             "str": SimpleIdAnalizer
#         }


# class DataLabelListAnalizer(Analizer):
#     def make_copy(self):
#         return DataLabelList(*self.data_labels)

#     @property
#     def data_labels(self):
#         return [self._analize_dataLabel(dataLabel) for dataLabel in self]

#     def __iter__(self):
#         yield from self.model.dataLabelList

#     def _analize_dataLabel(self, dataLabel):
#         return DataLabelAnalizer(dataLabel).make_copy()


# class DataLabelAnalizer(Analizer):
#     def make_copy(self):
#         return DataLabel(self.model)


# class StarAnalizer(Analizer):
#     def make_copy(self):
#         return All(self.model)


# class RefConstraintAnalizer(Analizer):
#     def make_copy(self):
#         return RefConstraint(self.model)





from fqllang.fql_stage.analizers.base import Analizer
from fqllang.fql_stage.analizers.create_form import CreateFormAnalizer
from fqllang.fql_stage.utils import getClassName


class ModelAnalizer(Analizer):
    def make_copy(self):
        className = getClassName(self.model)
        analizerDict = self.modelDict
        return analizerDict[className](self.model).make_copy()

    @property
    def modelDict(self):
        return {
            "CreateForm": CreateFormAnalizer
        }



from fqllang.fql_stage.ast_builders import FqlAstBuilder
from fqllang.fql_stage.utils import getGrammar
from fqllang.sql_stage.code_transpilers.postgreTranspiler import FqlAstPostgreSqlTranspiler

GRAMMAR = getGrammar()


def getPostgreCodeFromFqlCode(fqlCode):
    try:
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(fqlCode)
        postgreCode = FqlAstPostgreSqlTranspiler(fqlAstBuilder.model)
        return str(postgreCode).strip()
    except:
        raise Exception("Bad Fql Code")


if __name__ == "__main__":
    postgreCode = getPostgreCodeFromFqlCode("get Person with age=5 and name='Leandro'")
    print(postgreCode)
    postgreCode = getPostgreCodeFromFqlCode("get Person with age=5")
    print(postgreCode)
    postgreCode = getPostgreCodeFromFqlCode("get Person")
    print(postgreCode)
    postgreCode = getPostgreCodeFromFqlCode("create form Projects(title TEXT not null)")
    print(postgreCode)
    postgreCode = getPostgreCodeFromFqlCode("create form Employees(name TEXT, Project references 1..many Projects)")
    print(postgreCode)
    postgreCode = getPostgreCodeFromFqlCode("create form Person(name TEXT, Project references Projects)")
    print(postgreCode)
    postgreCode = getPostgreCodeFromFqlCode("create form Person(name TEXT, Project references 1 Projects)")
    print(postgreCode)
    postgreCode = getPostgreCodeFromFqlCode("create form Person(name TEXT, Project references 0..1 Projects)")
    print(postgreCode)
    postgreCode = getPostgreCodeFromFqlCode("create Project (name='Leandro', age=25)")
    print(postgreCode)
    



class SqlField:
    def __init__(self, name, datatype, *dataconstraints) -> None:
        self.name = name
        self.datatype = datatype.__type__
        self.dataconstraints = [dataconstraint.__type__ for dataconstraint in dataconstraints]
        
    def generate(self):
        datatype = self.datatype
        dataname_datatype = f"{self.name} {datatype}"
        dataconstraints = " ".join(self.dataconstraints)
        return f"{dataname_datatype} {dataconstraints}" if len(self.dataconstraints) else dataname_datatype

    def __str__(self) -> str:
        return self.generate()
    

class SqlForeignKeyField:
    def __init__(self, key, refpath, fieldName) -> None:
        self.key = key
        self.fieldName = fieldName
        self.refpath = refpath
        
    def generate(self):
        return f"foreign key({self.key}) references {self.refpath}({self.fieldName})"


from fqllang.sql_stage.components.dataconstraints import SqlDataconstraint, SqlNotnullDataconstraint, SqlPrimarykeyDataconstraint, SqlUniqueDataconstraint


class FqlDataconstraint:
    __type__ = "default"
    __sql__ = SqlDataconstraint


class FqlNotnullDataConstraint(FqlDataconstraint):
    __type__ = "not null"
    __sql__ = SqlNotnullDataconstraint


class FqlUniqueDataConstraint(FqlDataconstraint):
    __type__ = "unique"
    __sql__ = SqlUniqueDataconstraint


class FqlPrimarykeyDataConstraint(FqlDataconstraint):
    __type__ = "primary key"
    __sql__ = SqlPrimarykeyDataconstraint




from fqllang.fql_stage.ast_builders import FqlAstBuilder
from fqllang.sql_stage.code_transpilers.postgreTranspiler import FqlAstPostgreSqlTranspiler

def getSqlCode(fqlCode):
    fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(fqlCode)
    model = fqlAstBuilder.model
    postgreTanspiler = FqlAstPostgreSqlTranspiler(model)
    return postgreTanspiler.generate()
class TestFqlAstPostgreSqlTranspiler:
    def test_fqlStatementTranspiler_createForm(self):
        fqlCode = "create form Person(name TEXT, age NUMBER, project references Projects)"
        result = getSqlCode(fqlCode).lower()
        assert result == "CREATE TABLE Person(FQL_ID SERIAL PRIMARY KEY,name VARCHAR(511),age INTEGER,project INTEGER,FOREIGN KEY(project) REFERENCES Projects(FQL_ID));\n".lower()

    def test_fqlStatementTranspiler_getCase(self):
        fqlCode = "get Person"
        result = getSqlCode(fqlCode).lower()
        assert result == "SELECT * FROM Person;".lower()

    def test_fqlStatementTranspiler_getCaseWithConditions(self):
        fqlCode = "get Person with name='Leandro' and age=20"
        result = getSqlCode(fqlCode).lower()
        assert result == "SELECT * FROM Person WHERE name='Leandro' and age=20;".lower()
        
    def test_fqlStatementTranspiler_createNew(self):
        fqlCode = "create Person (name='Leandro', age=23)"
        result = getSqlCode(fqlCode).lower()
        assert result == "INSERT INTO Person (name,age) VALUES ('Leandro',23);".lower()




from fqllang.sql_stage.code_generators.insert import PreparedInsertPostgreSql
from fqllang.sql_stage.code_generators.basic_elements import ColumnList, ValueList


class TestPreparedInsertPostgreSql:
    def test_generate(self):
        tableName = "Person"
        columnList = ColumnList.createColumnListByElements('name', 'age')
        valueList = ValueList.createColumnListByElements('Leandro', 23)
        insert = PreparedInsertPostgreSql(tableName, columnList, valueList)
        assert insert.generate() == "INSERT INTO Person (name,age) VALUES ('Leandro',23);"



class SqlDataconstraint:
    __type__="default"
    
class SqlUniqueDataconstraint(SqlDataconstraint):
    __type__="unique"
    
class SqlNotnullDataconstraint(SqlDataconstraint):
    __type__="not null"
    
class SqlPrimarykeyDataconstraint(SqlDataconstraint):
    __type__="primary key"


from fqllang.core import getPostgreCodeFromFqlCode

postgreCode1 = getPostgreCodeFromFqlCode("create form Person (name TEXT, age NUMBER)")
postgreCode2 = getPostgreCodeFromFqlCode("get Person")
postgreCode3 = getPostgreCodeFromFqlCode("get Person with name='Leandro'")
postgreCode4 = getPostgreCodeFromFqlCode("get Person with name='Leandro' and age=20")

print(postgreCode1)
print(postgreCode2)
print(postgreCode3)
print(postgreCode4)



import os

def getGrammar():
    path = os.path.realpath(os.path.dirname(__file__))
    with open(f"{path}/grammar/fqlgrammar.tx") as f:
        return f.read()

def getGrammarPath():
    path = os.path.realpath(os.path.dirname(__file__))
    return f"{path}/grammar/fqlgrammar.tx"

def getClassName(object):
    return object.__class__.__name__


